/* 
   SquidTL - Time Limit for Squid proxy
   Copyright (C) 2009-2010, Michele Pinassi <o-zone@zerozone.it>.
   
   http://www.zerozone.it/Software/Linux/SquidTL
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "common.h"

FILE *logFHandle;

int logVerbosity;


#define CHAR_OK( c ) ( ((c)>31 && (c)<126) )

char *LOG_StripCR(char *str) {
    int iSrc, iDst;
    iSrc = iDst = 0;
    while (str[iSrc]) {
	if (CHAR_OK(str[iSrc])) {
    	    str[iDst++] = str[iSrc];
        }
        iSrc++;
    }
    str[iDst] = '\0'; 
    return str;
}

void LOG_Write(unsigned int level, const char* msg, ...) {
    va_list ap;
    char dateStr[20];
    struct tm tm;
    char levelStr[10],messageStr[8139];
    va_start(ap, msg);
    switch(level) {
	case DEBUG:
	    strcpy(levelStr,"DEBUG");
	    break;
	case NOTICE:
	    strcpy(levelStr,"NOTICE");
	    break;
	case WARNING:
	    strcpy(levelStr,"WARNING");
	    break;
	case ERROR:
	    strcpy(levelStr,"ERROR");
	    break;
	case FATAL:
	    strcpy(levelStr,"FATAL");
	    break;
	default:
	    strcpy(levelStr,"NOTICE");
    }
    time_t now = time(NULL);
    localtime_r(&now, &tm);
    strftime(dateStr, sizeof(dateStr), "%Y-%m-%d %H:%M:%S", &tm);
    vsprintf(messageStr, msg, ap);
    if(level >= logVerbosity) {
	fprintf(logFHandle, "[%d] %s %s - %s\n", getpid(), dateStr, levelStr, LOG_StripCR(messageStr));
	fflush(logFHandle);
    }
    va_end(ap);
}

void LOG_Init(char *logFile,int verbosity) {
    logFHandle = (FILE *)fopen(logFile,"a");
    if(logFHandle == NULL) {
        printf("LOG_Init(): Cannot open %s !\n",logFile);
	exit(1);
    }
    logVerbosity = verbosity;
    LOG_Write(FATAL,"LOG START (Verbosity: %d)",logVerbosity);
}

void LOG_Close() {
    LOG_Write(FATAL,"LOG CLOSE");
    fclose(logFHandle);
}
