/* 
   SquidTL - Time Limit for Squid proxy
   Copyright (C) 2009-2010, Michele Pinassi <o-zone@zerozone.it>.
   
   http://www.zerozone.it/Software/Linux/SquidTL
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "HTEscape.h"
#include <getopt.h>
#include "common.h"

#define MAX_BUF 4096

struct SquidInfo {
  char protocol[MAX_BUF];
  char domain[MAX_BUF];
  int  dot;  /* true if domain is in dot notation */
  char url[MAX_BUF];
  char orig[MAX_BUF];
  char surl[MAX_BUF];
  char furl[MAX_BUF];
  char *strippedurl;
  int  port;
  char src[MAX_BUF];
  char srcDomain[MAX_BUF];
  char ident[MAX_BUF];
  char method[MAX_BUF];
};

/* Converts a hex character to its integer value */
char from_hex(char ch) {
  return isdigit(ch) ? ch - '0' : tolower(ch) - 'a' + 10;
}

/* Converts an integer value to its hex character*/
char to_hex(char code) {
  static char hex[] = "0123456789abcdef";
  return hex[code & 15];
}

/* Returns a url-encoded version of str */
/* IMPORTANT: be sure to free() the returned string after use */
char *url_encode(char *str) {
  char *pstr = str, *buf = malloc(strlen(str) * 3 + 1), *pbuf = buf;
  while (*pstr) {
    if (isalnum(*pstr) || *pstr == '-' || *pstr == '_' || *pstr == '.' || *pstr == '~') 
      *pbuf++ = *pstr;
    else if (*pstr == ' ') 
      *pbuf++ = '+';
    else 
      *pbuf++ = '%', *pbuf++ = to_hex(*pstr >> 4), *pbuf++ = to_hex(*pstr & 15);
    pstr++;
  }
  *pbuf = '\0';
  return buf;
}


int parseLine(char *line, struct SquidInfo *s) {
  char *p, *d = NULL, *a = NULL, *e = NULL, *o, *field;
  int i = 0;
  char c;
  int report_once = 1;
  size_t strsz;
  int ndx = 0;
  
  field = strtok(line,"\t ");
  /*field holds each fetched url*/
  /* Let's first decode the url and then test it. Fixes bug2. */
  HTUnEscape(field);

  if(field == NULL)
    return 0;
  strcpy(s->orig,field);
  /* Now convert url to lowercase chars */ 
  for(p=field; *p != '\0'; p++) {
    *p = tolower(*p);
  }
  s->url[0] = s->protocol[0] = s->domain[0] = s->src[0] = s->ident[0] = 
    s->method[0] = s->srcDomain[0] = s->surl[0] =  '\0';
  s->dot = 0;
  s->port = 0;
  p = strstr(field,"://");
  if(p == NULL) { /* no protocol, defaults to http */
    strcpy(s->protocol,"unknown");
    p = field;
  } else {
    strncpy(s->protocol,field,p - field);
    *(s->protocol + ( p - field)) = '\0';
    p+=3; /* JMC -- 3 == strlen("://") */
    /* Now p only holds the pure URI */
    /* Fix for multiple slash vulnerability (bug1). */
    /* Check if there are still two or more slashes in sequence which must not happen */
    strsz = strlen(p);

    /* loop thru the string 'p' until the char '?' is hit or the "end" is hit */
    while('?' != p[ndx] && '\0' != p[ndx])
    {
        /* in case this is a '://' skip over it, but try to not read past EOS */
        if(3 <= strsz-ndx) {
          if(':' == p[ndx] && '/' == p[ndx+1] && '/' == p[ndx+2]) {
           ndx+=3; /* 3 == strlen("://"); */
          }
        }
        
       /* if this char and the next char are slashes,
 *           then shift the rest of the string left one char */
       if('/' == p[ndx] && '/' == p[ndx+1]) {
         size_t sz = strlen(p+ndx+1);
         strncpy(p+ndx,p+ndx+1, sz);
         p[ndx+sz] = '\0';
          if(1 == report_once) {
            // sgLogError("Warning: Possible bypass attempt. Found multiple slashes where only one is expected: %s", s->orig); 
            report_once--;
          }
      }
      else
      {
        /* increment the string indexer */
	assert(ndx < strlen(p));
        ndx++;
      }
    }
  }

  i=0;
  d = strchr(p,'/'); /* find domain end */
  /* Check for the single URIs (d) */
  e = d;
  a = strchr(p,'@'); /* find auth  */
  if(a != NULL && ( a < d || d == NULL)) 
    p = a + 1;
  a = strchr(p,':'); /* find port */;
  if(a != NULL && (a < d || d == NULL)){
    o = a + strspn(a+1,"0123456789") + 1;
    c = *o;
    *o = '\0';
    s->port = atoi(a+1);
    *o = c;
    e = a;
  }
  o=p;
  strcpy(s->furl,p);
  if (p[0] == 'w' || p[0] == 'f' ) {
    if ((p[0] == 'w' && p[1] == 'w' && p[2] == 'w') ||
	(p[0] == 'w' && p[1] == 'e' && p[2] == 'b') ||
	(p[0] == 'f' && p[1] == 't' && p[2] == 'p')) {
      p+=3;
      while (p[0] >= '0' && p[0] <= '9')
	p++;
      if (p[0] != '.')
	p=o; /* not a hostname */
      else
	p++;
    }
  }
  if(e == NULL){
    strcpy(s->domain,o);
    strcpy(s->surl,p);
  }
  else {
    strncpy(s->domain,o,e - o);
    strcpy(s->surl,p);
    *(s->domain + (e - o)) = '\0';
    *(s->surl + (e - p)) = '\0';
  }
  //strcpy(s->surl,s->domain);
  if(strspn(s->domain,".0123456789") == strlen(s->domain))
    s->dot = 1;
  if(d != NULL)
    strcat(s->surl,d);
  s->strippedurl = s->surl;

  while((p = strtok(NULL," \t\n")) != NULL){
    switch(i){
    case 0: /* src */
      o = strchr(p,'/');
      if(o != NULL){
	strncpy(s->src,p,o-p);
	strcpy(s->srcDomain,o+1);
	s->src[o-p]='\0';
	if(*s->srcDomain == '-') {
	  s->srcDomain[0] = '\0';
	}
      } else {
	strcpy(s->src,p);
      }
      break;
    case 1: /* ident */
      if(strcmp(p,"-")){
	strcpy(s->ident,p);
	for(p=s->ident; *p != '\0'; p++) /* convert ident to lowercase chars */
	  *p = tolower(*p);
      } else
	s->ident[0] = '\0';
      break;
    case 2: /* method */
      strcpy(s->method,p);
      break;
    }
    i++;
  }
  if(s->domain[0] == '\0') {
    LOG_Write(DEBUG,"Domain is NULL: %s", s->orig);
    return 0;
  }
  if(s->method[0] == '\0') {
    LOG_Write(DEBUG,"Method is NULL: %s", s->orig);
    return 0;
  }
  return 1;
}

unsigned int onShutdown = 0;

void Shutdown() {
    if(!onShutdown) {
	onShutdown = 1;
	LOG_Close();
	exit(0);
    }
}

int doRegExp(char *pattern, char *data) {
    regex_t *regex;
    int err_no=0,match=0; /* For regerror() */
    regmatch_t *result;

    
    regex = (regex_t *) malloc(sizeof(regex_t));
    memset(regex, 0, sizeof(regex_t));
    if((err_no=regcomp(regex, pattern, 0))!=0) /* Compile the regex */ {
	size_t length; 
	char *buffer;
	length = regerror (err_no, regex, NULL, 0);
	buffer = malloc(length);
	regerror(err_no, regex, buffer, length);
	LOG_Write(ERROR, "RegExp error: %s\n", buffer); /* Print the error */
	free(buffer);
	regfree(regex);
	return -1;
    }
    size_t no_sub = regex->re_nsub+1; /* How many matches are there in a line? */

    if((result = (regmatch_t *) malloc(sizeof(regmatch_t) * no_sub))==0) {
	LOG_Write(ERROR, "Cannot allocate required memory !"); /* Print the error */
	return -1;
    }
    
    if(regexec(regex, data, no_sub, result, 0) == 0) { /* Ok, match ! */
	match++;
    }

    regfree(regex); /* Free the regular expression data structure */
    free(regex);

    LOG_Write(DEBUG,"doRegExp: %s %s = %d\n",pattern,data,match);

    return match;
}

MYSQL *conn,*connb;
MYSQL_RES *res,*resb;
MYSQL_ROW row,rowb;

char buf[MAX_BUF],tmp[MAX_BUF];

void CB_Alarm(int sig) {
    LOG_Write(NOTICE,"CB_Alarm - Purge old entries\n");
    sprintf(tmp,"DELETE FROM DB_Access WHERE TIMESTAMPDIFF(MINUTE,NOW(),AddDate) > 360;");
    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
    if(mysql_query(conn, tmp)) {
        LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
    }
}

int main(int argc, char **argv) {
    struct SquidInfo squidInfo;
    time_t t;
    struct sigaction sact;

    char *redirect;

    char *server = "localhost";
    char *port = "3306";
    char *user = "squidtl";
    char *password = "secret";
    char *database = "squidtl";

    char *logfile = "/tmp/squidtl.log";
    
    char *defaultRedirectUrl = "http://proxy.sba.unisi.it/deny.php";

    char tmpEnc[4096];
    
    int rate=0,logVerbose=NOTICE,filterId;
    
    int c;

    struct option long_options[] = {
               {"verbose",     no_argument,       &logVerbose, DEBUG},
               {"help",        no_argument,       0, 'h'},
               {"mysql_server",required_argument, 0, 's'},
               {"mysql_port",  required_argument, 0, 't'},
               {"mysql_user",  required_argument, 0, 'u'},
               {"mysql_pwd",   required_argument, 0, 'p'},
               {"mysql_db",    required_argument, 0, 'd'},
               {"log_file",    required_argument, 0, 'l'},
               {"redirect_url",required_argument, 0, 'r'},
               {0, 0, 0, 0}
    };
     
    while (1) {
	/* getopt_long stores the option index here. */
	int option_index = 0;
	c = getopt_long (argc, argv, "hs:t:u:d:p:l:r:", long_options, &option_index);
	/* Detect the end of the options. */
	if(c == -1) break;
	switch (c) {
	    case 'h':
		printf("Usage: %s [options]\n\n--h -help  this help\n--s -mysql_server MySQL server (default: localhost)\n--t -mysql_port MySQL port (default: 3306)\n--u -mysql_user MySQL user (default: squidtl)\n--p -mysql_pwd MySQL password (default: secret)\n--d -mysql_db MySQL database (default: squidtl)\n--l -log_file Log file (default: %s)\n--r -redirect_url Default URL for blocking (default: %s)\n\n",logfile,defaultRedirectUrl);
		exit(0);
	    case 's':
		server = optarg;
		break;
	    case 't':
		port = optarg;
		break;
	    case 'u':
		user = optarg;
		break;
	    case 'p':
		password = optarg;
		break;
	    case 'd':
		database = optarg;
		break;
	    case 'l':
		logfile = optarg;
		break;
	    case 'r':
		defaultRedirectUrl = optarg;
		break;
	    default:
		break;
        }
    }
    
    LOG_Init(logfile,logVerbose);

    sigemptyset(&sact.sa_mask);
    sact.sa_flags = 0;
    sact.sa_handler = CB_Alarm;
    sigaction( SIGALRM, &sact, NULL );

    signal(SIGINT,  Shutdown); // Interrupt from keyboard (^C)
    signal(SIGHUP,  Shutdown); // Hangup detected on controlling terminal or death of controlling process
    signal(SIGQUIT, Shutdown); // Quit from keyboard (core dump)
    signal(SIGABRT, Shutdown); // Abort signal from abort(3)

    alarm(360);  /* Trigger ogni 360 secondi = 5 minuti */

    time(&t);

    conn = mysql_init(NULL);
    connb = mysql_init(NULL);

    /* Connect to database */
    if (!mysql_real_connect(conn, server, user, password, database, 0, NULL, 0)) {
	LOG_Write(ERROR,"MySQL Error: %s\n", mysql_error(conn));
	return(0);
    }

    if (!mysql_real_connect(connb, server, user, password, database, 0, NULL, 0)) {
	LOG_Write(ERROR,"MySQL Error: %s\n", mysql_error(conn));
	return(0);
    }

    LOG_Write(NOTICE,"MySQL(%s:%s) connection OK",server,user);

    while(1) {
	while(fgets(buf, MAX_BUF, stdin) != NULL) {
	    LOG_Write(DEBUG,"IN: %s",buf);
	    if(parseLine(buf,&squidInfo) != 1) {
		LOG_Write(ERROR,"Error parsing squid line: %s",buf);
		puts("");
	    } else {
		redirect = NULL;
		LOG_Write(DEBUG,"SquidInfo - P:%s D:%s U:%s O:%s S:%s F:%s\n",squidInfo.protocol, squidInfo.domain, squidInfo.url, squidInfo.orig, squidInfo.surl, squidInfo.furl);
		/* Add row */
		mysql_real_escape_string(conn, tmpEnc, squidInfo.orig, strlen(squidInfo.orig));
		sprintf(tmp,"INSERT INTO DB_Access (Src,Domain,FullUrl,Ident,AddDate) VALUES (\"%s\",\"%s\",\"%s\",\"%s\",NOW());",squidInfo.src,squidInfo.domain,tmpEnc,squidInfo.ident);
		LOG_Write(DEBUG,"MySQL: %s\n",tmp);
		if(mysql_query(conn, tmp)) {
		    LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
		}
		sprintf(tmp,"SELECT ID,Type,Value,Action,MaxHits,TimeSlice,redirectUrl FROM DB_Filter WHERE isActive=1;");
		/* 
		Type = Which type of rule is this ? (DOMAIN|IDENT|IP)
		Value = A value for this rule
		Action = What i have to do when triggered ? (BLOCK)
		MaxHits = How many times...
		TimeSlice = ...in how many minutes ?
		redirectUrl = Where i have to redirect the bad guys ?
		*/
		LOG_Write(DEBUG,"MySQL: %s\n",tmp);
		if(mysql_query(conn, tmp)) {
		    LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
		}
		res = mysql_use_result(conn);
		while((row = mysql_fetch_row(res)) != NULL) {
		    /* === DOMAIN LIMITER === */
		    if((strcasecmp(row[1],"DOMAIN") == 0) && (doRegExp(row[2],squidInfo.domain) > 0)) {
			/* Ok, now check the value */
			sprintf(tmp,"SELECT Domain FROM DB_Access WHERE Ident=\"%s\" AND TIMESTAMPDIFF(MINUTE,AddDate,NOW()) < %s;",squidInfo.ident,row[5]);
			LOG_Write(DEBUG,"MySQL: %s\n",tmp);
			if(mysql_query(connb, tmp)) {
			    LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(connb));
			} else {
			    resb = mysql_use_result(connb);
			    rate = 0;
			    while((rowb = mysql_fetch_row(resb)) != NULL) {
				if(doRegExp(row[2],rowb[0]) > 0) {
				    rate++;
				}
			    }
			    LOG_Write(DEBUG,"LIMIT for domain %s: %s hits in %s minutes. Ident %s have %d\n",row[2],row[4],row[5],squidInfo.ident,rate);
			    if(rate > atoi(row[4])) {
				char *redirUrl, *urlencode;
				if(strlen(row[6]) < 1) {
				    redirUrl = defaultRedirectUrl;
				} else {
				    redirUrl = row[6];
				}
				urlencode = url_encode(row[2]);
				sprintf(tmp,"%s?rule=domain&action=%s&domain=%s&ident=%s&hits=%d&hitlimit=%s&timelimit=%s",redirUrl,row[3],urlencode,squidInfo.ident,rate,row[4],row[5]);
				free(urlencode);
				redirect = tmp;
				filterId = atoi(row[0]);
			    }
			    mysql_free_result(resb);
			}
		    } 
		    /* === IDENT LIMITER === */
		    if((strcasecmp(row[1],"IDENT") == 0) && (doRegExp(row[2],squidInfo.ident) > 0)) {
			/* Ok, now check the value */
			sprintf(tmp,"SELECT COUNT(ID) FROM DB_Access WHERE Ident=\"%s\" AND TIMESTAMPDIFF(MINUTE,AddDate,NOW()) < %s;",squidInfo.ident,row[5]);
			LOG_Write(DEBUG,"MySQL: %s\n",tmp);
			if(mysql_query(connb, tmp)) {
			    LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(connb));
			} else {
			    resb = mysql_use_result(connb);
			    rowb = mysql_fetch_row(resb);
			    rate = atoi(rowb[0]);
			    LOG_Write(DEBUG,"LIMIT for ident %s: %s hits in %s minutes. Now we have %d\n",squidInfo.ident,row[4],row[5],rate);
			    if(rate > atoi(row[4])) {
				char *redirUrl;
				if(strlen(row[6]) < 1) {
				    redirUrl = defaultRedirectUrl;
				} else {
				    redirUrl = row[6];
				}
				sprintf(tmp,"%s?rule=ident&action=%s&ident=%s&hits=%d&hitlimit=%s&timelimit=%s",redirUrl,row[3],squidInfo.ident,rate,row[4],row[5]);
				redirect = tmp;
				filterId = atoi(row[0]);
			    }
			    mysql_free_result(resb);
			}
		    } 
		    /* === IP LIMITER === */
		    if((strcasecmp(row[1],"IP") == 0) && (doRegExp(row[2],squidInfo.src) > 0)) {
			/* Ok, now check the value */
			sprintf(tmp,"SELECT COUNT(ID) FROM DB_Access WHERE Src=\"%s\" AND TIMESTAMPDIFF(MINUTE,AddDate,NOW()) < %s;",squidInfo.src,row[5]);
			LOG_Write(DEBUG,"MySQL: %s\n",tmp);
			if(mysql_query(connb, tmp)) {
			    LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(connb));
			} else {
			    resb = mysql_use_result(connb);
			    rowb = mysql_fetch_row(resb);
			    rate = atoi(rowb[0]);
			    LOG_Write(DEBUG,"LIMIT for IP %s: %s hits in %s minutes. Now we have %d\n",squidInfo.src,row[4],row[5],rate);
			    if(rate > atoi(row[4])) {
				char *redirUrl;
				if(strlen(row[6]) < 1) {
				    redirUrl = defaultRedirectUrl;
				} else {
				    redirUrl = row[6];
				}
				sprintf(tmp,"%s?rule=ip&action=%s&ip=%s&hits=%d&hitlimit=%s&timelimit=%s",redirUrl,row[3],squidInfo.src,rate,row[4],row[5]);
				redirect = tmp;
				filterId = atoi(row[0]);
			    }
			    mysql_free_result(resb);
			}
		    }
		}
		/* Now check for triggers */
		if(redirect != NULL) {
		    /* Oh oh, bad guy...needs to be redirected ! */
		    fprintf(stdout,"302:%s\n",redirect);
		    LOG_Write(NOTICE,"REDIRECT: 302:%s\n",redirect);
		    /* Save event to DB for statistic purposes */
		    sprintf(tmp,"INSERT INTO DB_Events (FilterID,Src,Ident,AddDate) VALUES (%d,\"%s\",\"%s\",NOW());",filterId,squidInfo.src,squidInfo.ident);
		    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
		    if(mysql_query(conn, tmp)) {
			LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
		    }
		} else {
		    puts("");
		}
		mysql_free_result(res);
		break;
	    }
	}
	fflush(stdout);
    }
    mysql_close(connb);
    mysql_close(conn);
}
